/* DChub - a Direct Connect Hub for Linux
 * Copyright (C) 2001 Eric Prevoteau
 *
 * hub_cmd.c: Copyright (C) Eric Prevoteau <www@ac2i.tzo.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <glib.h>
#include "cnx_lst.h"
#include "hub_cmd.h"
#include "gvar.h"
#include "macro.h"
#include "passwd.h"
#include "toolkit.h"
#include "e_db.h"
#include "multi_private_chat.h"
#include "timed_out_string.h"
#include "tos_key.h"
#include "main.h"
#include "hub_cluster.h"
#include "multi_public_chat.h"

static int do_reload(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	load_dyn_conf();
	g_string_sprintfa((*dest_base),"Reload done|");
	send_const_str_to(ptr,(*dest_base)->str);
	return (0);
}

#ifdef DEBUG

static int do_dump(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{

	if(splitted_str->len<2)		/* at least 1 parameters + the command */
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
		send_const_str_to(ptr,(*dest_base)->str);
	}
	else
	{
		int i;
		const char *user;
		CNX_ENTRY *ptr2;

		for(i=1; i<splitted_str->len; i++)
		{
			user=g_ptr_array_index(splitted_str,i);
			if(user!=NULL)
			{
				GString *tmp;
#ifdef __CYGWIN__
				char		long_int[48];
#endif		
				ptr2 = find_cnx_by_nickname(user); 
				tmp = g_string_new((*dest_base)->str);
				
				if (ptr2 == NULL)
					g_string_sprintfa(tmp, "nick : %s not found in strucure|", user);
				else
				{
#ifdef __CYGWIN__
					sprintf(long_int, "%Lu", ptr2->shared_size);
					g_string_sprintfa(tmp, "user_nick=%s, sock_fd=%d, user_cnx_type=%s, is_op=%i, IP=%s, share=%s, incomming=%p, outcomming=%p, server_id=%d, local=%p, next=%p|",
#else
					g_string_sprintfa(tmp, "user_nick=%s, sock_fd=%d, user_cnx_type=%s, is_op=%i, IP=%s, share=%Lu, incomming=%p, outcomming=%p, server_id=%d, local=%p, next=%p|",
#endif					
									ptr2->user_nick->str,
									ptr2->sock_fd,
									ptr2->user_cnx_type,
									ptr2->is_op,
									inet_ntoa(ptr2->ip),
#ifdef __CYGWIN__
									long_int,
#else
									ptr2->shared_size,
#endif
									ptr2->incoming_commands,
									ptr2->outgoing_commands,
									ptr2->server_id,
									ptr2,
									ptr2->next);
				}
				send_const_str_to(ptr,tmp->str);
				g_string_free(tmp, TRUE);
			}
		}
	}
	return (0);
}

static int do_dump_queue(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	CNX_ENTRY	 *ptr2;
	GString	*tmp;

	for (ptr2=cnx_lst->wait; ptr2; ptr2 = ptr2->next)
	{
		tmp = g_string_new((*dest_base)->str);
		if (ptr2->handshake_done==1)
		{
			g_string_sprintfa(tmp,"user_nick=%s, sock_fd=%d, user_cnx_type=%s, is_op=%i, IP=%s, incomming=%p, outcomming=%p, server_id=%d, local=%p, next=%p|",
						ptr2->user_nick->str,
						ptr2->sock_fd,
						ptr2->user_cnx_type,
						ptr2->is_op,
						inet_ntoa(ptr2->ip),
						ptr2->incoming_commands,
						ptr2->outgoing_commands,
						ptr2->server_id,
						ptr2,
						ptr2->next);
		}
		else
		{
			g_string_sprintfa(tmp,"%duser_nick=?????, sock_fd=%d, user_cnx_type=%s, is_op=%i, IP=%s, incomming=%p, outcomming=%p, server_id=%d, local=%p, next=%p|",ptr2->connected,
						ptr2->sock_fd,
						ptr2->user_cnx_type,
						ptr2->is_op,
						inet_ntoa(ptr2->ip),
						ptr2->incoming_commands,
						ptr2->outgoing_commands,
						ptr2->server_id,
						ptr2,
						ptr2->next);
		}
		send_const_str_to(ptr,tmp->str);
		g_string_free(tmp, TRUE);
	}
	return(0);
}

#endif

static int do_set_fusr(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	if(splitted_str->len!=2)		/* at least 1 parameters + the command */
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
		send_const_str_to(ptr,(*dest_base)->str);
	}
	else if (route_table->server_count > 1)
	{
		g_string_sprintfa((*dest_base),"the domain name can not be changed once a group had been create, "
												 "there is currently %d hubs in the group\r\n|",route_table->server_count);
		send_const_str_to(ptr,(*dest_base)->str);
	}
	else
	{
		if (fusion_user != NULL)
			free(fusion_user);
		fusion_user = strdup(g_ptr_array_index(splitted_str,1));
	}
	return(0);
}

static int do_set_fpass(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	if(splitted_str->len!=2)		/* at least 1 parameters + the command */
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
		send_const_str_to(ptr,(*dest_base)->str);
	}
	else if (route_table->server_count > 1)
	{
		g_string_sprintfa((*dest_base),"the domain passwd can not be changed once a group had been create, "
												 "there is currently %d hubs in the group\r\n|", route_table->server_count);
		send_const_str_to(ptr,(*dest_base)->str);
	}
	else
	{
		if (fusion_pass != NULL)
			free(fusion_pass);
		fusion_pass = strdup(g_ptr_array_index(splitted_str,1));
	}
	return(0);
}

static int do_fusion(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	if(splitted_str->len!=2)		/* at least 1 parameters + the command */
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
	else if ((fusion_pass == NULL) || (fusion_user == NULL))
		g_string_sprintfa((*dest_base),"fusion user/password not set\r\n|");
	else if (interhub_cnx(g_ptr_array_index(splitted_str,1), 1))
		g_string_sprintfa((*dest_base),"fusion failure : no connection to remote hub\r\n|");
	else
		g_string_sprintfa((*dest_base),"WARNING incoming connection\r\n|");
	send_const_str_to(ptr,(*dest_base)->str);
	return(0);
}

static int do_unfusion(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	int	i;

	for (i = 1; i <= route_table->max_server_id; i++)
	{
		if (route_table->table[i].connected != 0)
			lost_hub(route_table->table + i);
	}
	return(0);
}

static int do_set(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	if(splitted_str->len!=3)
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s. Usage : set var value\r\n|",cmd);
	else if (!strcmp(g_ptr_array_index(splitted_str,1), "MODERATE"))
	{
		printf("(%s)", (char *)g_ptr_array_index(splitted_str,2));
		sscanf(g_ptr_array_index(splitted_str,2), "%d", &gl_moderate_mode);
		if (gl_moderate_mode)
			g_string_sprintfa((*dest_base)," moderate mode enabled\r\n|");
		else
			g_string_sprintfa((*dest_base)," moderate mode disabled\r\n|");
	}
	g_string_sprintfa((*dest_base),"arg 1 in MODERATE\r\n|");
	send_const_str_to(ptr,(*dest_base)->str);
	return(0);
}

/****************************/
/* get ip from a named user */
/****************************/

static int do_get_ip(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	if(splitted_str->len != 2)		/* at least 1 parameters + the command */
	{
		g_string_sprintfa((*dest_base),"usage: %s user_name\r\n|",cmd);
	}
	else
	{
		const char *user;
		CNX_ENTRY *ptr2;

		user=g_ptr_array_index(splitted_str,1);

		if(user!=NULL)
		{
			ptr2 = find_cnx_by_nickname(user);
			if (ptr2)
			{
				g_string_sprintfa((*dest_base),"ip:%s %s\r\n|", ptr2->user_nick->str, inet_ntoa(ptr2->ip));
			}
			else
			{
				g_string_sprintfa((*dest_base),"user %s not found.\r\n|", user);
			}
		}
	}
	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}

/****************************************/
/* give the username having the same IP */
/**********************************-*****/

static int do_rev_ip(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	if(splitted_str->len != 2)		/* at least 1 parameters + the command */
	{
		g_string_sprintfa((*dest_base),"usage: %s user_name\r\n|",cmd);
	}
	else
	{
		const char *str_ip;
		str_ip=g_ptr_array_index(splitted_str,1);

		if(str_ip!=NULL)
		{
			int i;
			unsigned long int ip;

			if ((ip = inet_addr(str_ip)) == INADDR_NONE)
				g_string_sprintfa((*dest_base),"unable to parse %s\r\n|", str_ip);
			else
			{
				GString *out;
				out=g_string_new("");
				for(i=0;i<HASH_SIZE;i++)
				{
					CNX_ENTRY *ptr2;

					for (ptr2=cnx_lst->entry[i]; ptr2; ptr2=ptr2->next)
					{
						if (ptr2->ip.s_addr== ip)
						{
							out=g_string_assign(out,(*dest_base)->str);
							g_string_sprintfa(out,"rev:%s %s|",inet_ntoa(ptr2->ip), ptr2->user_nick->str);
							send_const_str_to(ptr,out->str);
						}
					}
				}
				g_string_free(out,TRUE);
				/* g_string_free((*dest_base),TRUE); */
				/* (*dest_base)=0; */
				return(0);
			}
		}
		else
		{
			g_string_sprintfa((*dest_base),"internal error in %s\r\n|", __FILE__);
		}
	} 
	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}

/************************************/
/* display the list of all commands */
/************************************/

static int do_help(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	g_string_sprintfa((*dest_base),"** Operator command list:\r\n"
											 "\r\n"
											 "-help                         display this help\r\n"
											 "+help                         normal user help\r\n"
										 	 "\r\n"
											 "-kick nick reason             kick the given user out of the hub\r\n"
											 "-skick nick reason            silent kick (Without printing anything on the chat)\r\n"
											 "-lkick nick reason            disconnect only\r\n"
											 "-bankick nick reason          kick and permban a user\r\n"
											 "-redir nick address           redirect the given user to this address\r\n"
											 "-bcast msg                    send the given message to everyone\r\n"
											 "\r\n"
											 "-getip nick                   Get IP of a user.\r\n"
											 "-revip ip                     Reveal's all the users having the given IP.\r\n"
											 "\r\n"
											 "-tbanip ip [ip...]            add temporary ban on the given ip(s)\r\n"
											 "-tunbanip ip [ip...]          remove temporary ban of the given ip(s)\r\n"
											 "-tbanlist                     print the list of temporary banned ip(s)\r\n"
											 "-pbanip ip [ip...]            add permanent ban on the given ip(s)\r\n"
											 "-punbanip ip [ip...]          remove permanent ban of the given ip(s)\r\n"
											 "-pbanlist                     print the list of permanently banned ip(s)\r\n"
											 "\r\n"
											 "\r\n"
											 "*** Master command list:\r\n"
											 "\r\n"
											 "-ulist		display all existing accounts\r\n"
											 "\r\n"
											 "-add nick passwd type         create a new registered user\r\n"
											 "-del nick                     delete a new registered user\r\n"
											 "-rename oldnick newnick       rename a registered user\r\n"
											 "-passwd nick newpasswd        change the password of a registered user\r\n"
											 "-type nick newtype            change the type of a registered user\r\n"
											 "\r\n"
											 "-db                           display database\r\n"
											 "-dbshow key                   display one database key\r\n"
											 "-dbadd key type value         create a new key in the database\r\n"
											 "-dbset key value              change the value of a key in the database\r\n"
											 "-dbdel key                    delete a key of the database\r\n"
							 				 "-reload                       update some variables in the hub\r\n"
											 "\r\n"
							 				 "-set VARIABLE VALUE           custom function (new)\r\n"
											 "-startprg name                start the given external program (-e required)\r\n"
											 "                              during hub start and name must be a valid name\r\n"
											 "                              in this directory\r\n"
											 "\r\n"
											 "-redirall address             redirect all users to this address\r\n"
							 				 "-rediralllocal address        redirect all local users to this address\r\n"
#ifdef DEBUG
							 				 "-dump user [users...]         dump data from the internal structure (use with care)\r\n"
							 				 "-dumpqueue                    dump login queue list (use with care)\r\n"
#endif
							 				 "\r\n"
							 				 "\r\n"
							 				 "*** Fusion command list:\r\n"
							 				 "-setfusionuser name           set the fusion username\r\n"
							 				 "-setfusionpass password       set the fusion password\r\n"
							 				 "-fusion ip:port               fusion with a hub group. EXPERIMENTAL\r\n"
							 				 "                              the other hubs must have the same fusion user/password\r\n"
							 				 "-ufusion                      disconnect from the cluster\r\n"
											 "\r\n"
											 "\r\n"
											 "Note: If a parameter contains spaces, put the parameter between quotes\r\n"
											 "Note2: A quoted parameter cannot contains quote itself\r\n"
											 "\r\n|");
	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}

/***************************************************************************************/
/* display the list of all existing account with their account type and their password */
/***************************************************************************************/
static int do_ulist(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	GString *out;

	out=full_user_list();
	g_string_sprintfa((*dest_base),"%s|",out->str);
	g_string_free(out,TRUE);
	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}

/*******************************************************/
/* add new registered user to the registered user list */
/*******************************************************/
static int do_add(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	if(splitted_str->len!=4)		/* 3 parameters + the command */
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
	}
	else
	{
		char *t;

		/* check login */
		t=g_ptr_array_index(splitted_str,1);
		if((strlen(t)+1)>NICK_PASS_LEN)
		{
			g_string_sprintfa((*dest_base),"%s: Nickname too long (max: %d).\r\n|",cmd,NICK_PASS_LEN-1);
			goto leave;
		}

		/* account already exists ? */
		if(account_exist(t,NULL))
		{
			g_string_sprintfa((*dest_base),"%s: an account with this nickname still exists.\r\n|",cmd);
			goto leave;
		}

		/* check password */
		t=g_ptr_array_index(splitted_str,2);
		if((strlen(t)+1)>PASS_PASS_LEN)
		{
			g_string_sprintfa((*dest_base),"%s: Password too long (max: %d).\r\n|",cmd,PASS_PASS_LEN-1);
			goto leave;
		}

		/* check type */
		t=g_ptr_array_index(splitted_str,3);
		if(strlen(t)!=1)
		{
			g_string_sprintfa((*dest_base),"%s: Account type is only 1 character (N, O, M, b, B or F).\r\n|",cmd);
			goto leave;
		}

		if((t[0]!='N')&&(t[0]!='O')&&(t[0]!='M')&&(t[0]!='b')&&(t[0]!='B')&&(t[0]!='F'))
		{
			g_string_sprintfa((*dest_base),"%s: Account type can only be N, O, M, b, B or F.\r\n|",cmd);
			goto leave;
		}

		if(add_user_to_ulist(g_ptr_array_index(splitted_str,1),g_ptr_array_index(splitted_str,2),*t)!=0)
		{
			g_string_sprintfa((*dest_base),"%s: Error while creating nickname %s.\r\n|",cmd,(char*)g_ptr_array_index(splitted_str,1));
		}
		else
		{
			g_string_sprintfa((*dest_base),"%s: nickname %s created.\r\n|",cmd,(char*)g_ptr_array_index(splitted_str,1));
		}
	}
	leave:

	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}

/**********************************************************/
/* delete a registered user from the registered user list */
/**********************************************************/
static int do_del(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	if(splitted_str->len!=2)		/* 1 parameter + the command */
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
	}
	else
	{
		char *t;

		/* check login */
		t=g_ptr_array_index(splitted_str,1);
		if((strlen(t)+1)>NICK_PASS_LEN)
		{
			g_string_sprintfa((*dest_base),"%s: Nickname too long (max: %d).\r\n|",cmd,NICK_PASS_LEN-1);
			goto leave;
		}

		/* account already exists ? */
		if(!account_exist(t,NULL))
		{
			g_string_sprintfa((*dest_base),"%s: No account with this nickname exists.\r\n|",cmd);
			goto leave;
		}

		if(del_user_to_ulist(g_ptr_array_index(splitted_str,1))!=0)
		{
			g_string_sprintfa((*dest_base),"%s: Error while deleting nickname %s.\r\n|",cmd,(char*)g_ptr_array_index(splitted_str,1));
		}
		else
		{
			g_string_sprintfa((*dest_base),"%s: nickname %s deleted.\r\n|",cmd,(char*)g_ptr_array_index(splitted_str,1));
		}
	}
	leave:

	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}

/******************************************************************/
/* rename an existing registered user of the registered user list */
/******************************************************************/
static int do_rename(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	if(splitted_str->len!=3)		/* 2 parameters + the command */
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
	}
	else
	{
		char *t;

		/* check old login */
		t=g_ptr_array_index(splitted_str,1);
		if((strlen(t)+1)>NICK_PASS_LEN)
		{
			g_string_sprintfa((*dest_base),"%s: Old Nickname too long (max: %d).\r\n|",cmd,NICK_PASS_LEN-1);
			goto leave;
		}

		/* account already exists ? */
		if(!account_exist(t,NULL))
		{
			g_string_sprintfa((*dest_base),"%s: No account with this nickname exists.\r\n|",cmd);
			goto leave;
		}

		/* check new login */
		t=g_ptr_array_index(splitted_str,2);
		if((strlen(t)+1)>NICK_PASS_LEN)
		{
			g_string_sprintfa((*dest_base),"%s: New Nickname too long (max: %d).\r\n|",cmd,NICK_PASS_LEN-1);
			goto leave;
		}

		/* account already exists ? */
		if(account_exist(t,NULL))
		{
			g_string_sprintfa((*dest_base),"%s: An account having the new name still exists.\r\n|",cmd);
			goto leave;
		}

		if(rename_user_to_ulist(g_ptr_array_index(splitted_str,1),g_ptr_array_index(splitted_str,2))!=0)
		{
			g_string_sprintfa((*dest_base),"%s: Error while updating nickname %s.\r\n|",cmd,(char*)g_ptr_array_index(splitted_str,1));
		}
		else
		{
			g_string_sprintfa((*dest_base),"%s: nickname %s updated.\r\n|",cmd,(char*)g_ptr_array_index(splitted_str,1));
		}
	}
	leave:

	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}

/******************************************************/
/* change the password of an existing registered user */
/******************************************************/
static int do_passwd(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	if(splitted_str->len!=3)		/* 2 parameters + the command */
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
	}
	else
	{
		char *t;

		/* check old login */
		t=g_ptr_array_index(splitted_str,1);
		if((strlen(t)+1)>NICK_PASS_LEN)
		{
			g_string_sprintfa((*dest_base),"%s: Nickname too long (max: %d).\r\n|",cmd,NICK_PASS_LEN-1);
			goto leave;
		}

		/* account already exists ? */
		if(!account_exist(t,NULL))
		{
			g_string_sprintfa((*dest_base),"%s: No account with this nickname exists.\r\n|",cmd);
			goto leave;
		}

		/* check password */
		t=g_ptr_array_index(splitted_str,2);
		if((strlen(t)+1)>PASS_PASS_LEN)
		{
			g_string_sprintfa((*dest_base),"%s: Password too long (max: %d).\r\n|",cmd,PASS_PASS_LEN-1);
			goto leave;
		}

		if(chg_user_passwd_to_ulist(g_ptr_array_index(splitted_str,1),g_ptr_array_index(splitted_str,2))!=0)
		{
			g_string_sprintfa((*dest_base),"%s: Error while updating password of nickname %s.\r\n|",cmd,(char*)g_ptr_array_index(splitted_str,1));
		}
		else
		{
			g_string_sprintfa((*dest_base),"%s: password of nickname %s updated.\r\n|",cmd,(char*)g_ptr_array_index(splitted_str,1));
		}
	}
	leave:

	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}

/*******************************************/
/* change an existing registered user type */
/*******************************************/
static int do_type(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	if(splitted_str->len!=3)		/* 2 parameters + the command */
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
	}
	else
	{
		char *t;

		/* check old login */
		t=g_ptr_array_index(splitted_str,1);
		if((strlen(t)+1)>NICK_PASS_LEN)
		{
			g_string_sprintfa((*dest_base),"%s: Nickname too long (max: %d).\r\n|",cmd,NICK_PASS_LEN-1);
			goto leave;
		}

		/* account already exists ? */
		if(!account_exist(t,NULL))
		{
			g_string_sprintfa((*dest_base),"%s: No account with this nickname exists.\r\n|",cmd);
			goto leave;
		}

		/* check type */
		t=g_ptr_array_index(splitted_str,2);
		if(strlen(t)!=1)
		{
			g_string_sprintfa((*dest_base),"%s: Account type is only 1 character (N, O, M, b, B or F).\r\n|",cmd);
			goto leave;
		}

		if ((t[0]!='N')&&(t[0]!='O')&&(t[0]!='M')&&(t[0]!='b')&&(t[0]!='B')&&(t[0]!='F'))
		{
			g_string_sprintfa((*dest_base),"%s: Account type can only be N, O, M, b, B or F.\r\n|",cmd);
			goto leave;
		}

		if(chg_user_type_to_ulist(g_ptr_array_index(splitted_str,1),*t)!=0)
		{
			g_string_sprintfa((*dest_base),"%s: Error while updating type of nickname %s.\r\n|",cmd,(char*)g_ptr_array_index(splitted_str,1));
		}
		else
		{
			g_string_sprintfa((*dest_base),"%s: type of nickname %s updated.\r\n|",cmd,(char*)g_ptr_array_index(splitted_str,1));
		}
	}
	leave:

	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}

/*************************/
/* dump current database */
/*************************/
static int do_db(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	if(splitted_str->len!=1)		/* 0 parameters + the command */
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
	}
	else
	{
		GStringChunk *gsc=NULL;
		GPtrArray *gpa_name=NULL;
		GPtrArray *gpa_type=NULL;
		GPtrArray *gpa_val=NULL;

		if(!get_all_db_keys(&gpa_name,&gpa_type,&gpa_val,&gsc))
		{
			int i;
			GString *out;

			out=g_string_new("");

			/* the database content is sent in multiple small messages */
			/* to avoid buffer overflow in DC */
			for(i=0;i<gpa_name->len;i++)
			{
				out=g_string_assign(out,(*dest_base)->str);
				g_string_sprintfa(out,"'%s'\t'%s'\t'%s'|",
											(char*)g_ptr_array_index(gpa_type,i),
											(char*)g_ptr_array_index(gpa_name,i),
											(char*)g_ptr_array_index(gpa_val,i));
				send_const_str_to(ptr,out->str);
			}
			g_string_free(out,TRUE);
		}
		
		if(gsc!=NULL)
			g_string_chunk_free(gsc);
		if(gpa_name!=NULL)
			g_ptr_array_free(gpa_name,TRUE);
		if(gpa_type!=NULL)
			g_ptr_array_free(gpa_type,TRUE);
		if(gpa_val!=NULL)
			g_ptr_array_free(gpa_val,TRUE);

		g_string_sprintfa((*dest_base),"End of database dump|");
	}
	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}

/**********************/
/* dump the key value */
/**********************/
static int do_dbshow(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	if(splitted_str->len!=2)		/* 1 parameters + the command */
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
	}
	else
	{
		char *tp;
		char *kn;

		G_LOCK(conf_file);
		kn=g_ptr_array_index(splitted_str,1);

		tp=e_db_type_get(conf_file,kn);
		if(tp==NULL)
		{
			g_string_sprintfa((*dest_base),"no key named '%s'\r\n|",kn);
		}
		else
		{
			if(!strcmp(tp,"int"))
			{
				/* value is an int */
				int vl;

				if(!e_db_int_get(conf_file,kn,&vl))
					vl=0;

				g_string_sprintfa((*dest_base),"'%s'\t'%s'\t'%d'|",
											tp,kn,vl);
			}
			else if(!strcmp(tp,"float"))
			{
				/* value is a float */
				float vl;

				if(!e_db_float_get(conf_file,kn,&vl))
					vl=0;

				g_string_sprintfa((*dest_base),"'%s'\t'%s'\t'%f'|",
											tp,kn,vl);
			}
			else
			{
				/* value is a string */
				char *vl;

				vl=e_db_str_get(conf_file,kn);
				if(vl==NULL)
				{
					g_string_sprintfa((*dest_base),"'%s'\t'%s'\t''|", tp,kn);
				}
				else
				{
					g_string_sprintfa((*dest_base),"'%s'\t'%s'\t'%s'|",
											tp,kn,vl);
					free(vl);
				}
			}
			free(tp);
		}
		G_UNLOCK(conf_file);
	}

	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}

/*********************/
/* add the key value */
/*********************/
static int do_dbadd(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	if(splitted_str->len!=4)		/* 3 parameters + the command */
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
	}
	else
	{
		char *tp;
		char *kn;

		G_LOCK(conf_file);

		kn=g_ptr_array_index(splitted_str,1);

		tp=e_db_type_get(conf_file,kn);
		if(tp!=NULL)
		{
			g_string_sprintfa((*dest_base),"a key with this name still exists. Error\r\n|");
			free(tp);
		}
		else
		{
			/* check type before adding */
			tp=g_ptr_array_index(splitted_str,2);
			if((strcmp(tp,"int")) && (strcmp(tp,"str")) && (strcmp(tp,"float")))
			{
				g_string_sprintfa((*dest_base),"Unknown type. Only 'int', 'str' and 'float' are allowed. Error\r\n|");
			}
			else
			{
				if(!strcmp(tp,"int"))
				{
					e_db_int_set(conf_file,kn,atoi(g_ptr_array_index(splitted_str,3)));
				}
				else if(!strcmp(tp,"str"))
				{
					e_db_str_set(conf_file,kn,g_ptr_array_index(splitted_str,3));
				}
				else
				{
					e_db_float_set(conf_file,kn,atof(g_ptr_array_index(splitted_str,3)));
				}
				g_string_sprintfa((*dest_base),"Key %s added.\r\n|",kn);
			}
		}
		G_UNLOCK(conf_file);
	}

	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}

/************************/
/* update the key value */
/************************/
static int do_dbset(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	if(splitted_str->len!=3)		/* 2 parameters + the command */
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
	}
	else
	{
		char *tp;
		char *kn;

		G_LOCK(conf_file);

		kn=g_ptr_array_index(splitted_str,1);

		tp=e_db_type_get(conf_file,kn);
		if(tp==NULL)
		{
			g_string_sprintfa((*dest_base),"No key with this name still exists. Error\r\n|");
		}
		else
		{
			if(!strcmp(tp,"int"))
			{
				e_db_int_set(conf_file,kn,atoi(g_ptr_array_index(splitted_str,2)));
			}
			else if(!strcmp(tp,"str"))
			{
				e_db_str_set(conf_file,kn,g_ptr_array_index(splitted_str,2));
			}
			else
			{
				e_db_float_set(conf_file,kn,atof(g_ptr_array_index(splitted_str,2)));
			}
			g_string_sprintfa((*dest_base),"Key %s updated.\r\n|",kn);
			free(tp);
		}
		G_UNLOCK(conf_file);
		load_dyn_conf();
	}
	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}

/********************************************/
/* test if the given key is a protected key */
/********************************************/
/* out: 0=not protected, 1=yes */
static int is_a_protected_key(char *kn)
{
	static const char *tbl_prot[]={
												"HUBNAME",
												"HUBADDR",
												"HUBDESC",
												"REG_HUB",
												"REG_SERV",
												"MAXUSER",
												"REDIR_FULL",
												"REDIR_ADDR",
												"ONLY_REG",
												NULL
											};
	int i;
	
	i=0;
	while(tbl_prot[i]!=NULL)
	{
		if(!strcmp(kn,tbl_prot[i]))
			return 1;
		i++;
	}

	return 0;
}

/*********************/
/* del the key value */
/*********************/
static int do_dbdel(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	if(splitted_str->len!=2)		/* 1 parameters + the command */
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
	}
	else
	{
		char *kn;

		kn=g_ptr_array_index(splitted_str,1);
		if(!is_a_protected_key(kn))
		{
			e_db_data_del(conf_file,kn);
			g_string_sprintfa((*dest_base),"Key %s deleted.\r\n|",kn);
		}
		else
		{
			g_string_sprintfa((*dest_base),"Key %s is a protected key, you cannot delete it.\r\n|",kn);
		}
	}

	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}

/*****************************/
/* start an external program */
/*****************************/
static int do_startprg(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	char *progname;
	gchar *p;
	struct stat st;

	if(splitted_str->len!=2)		/* 1 parameters + the command */
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
		goto abrt;
	}

	if(ext_prog_dir==NULL)
	{
		g_string_sprintfa((*dest_base),"Hub has been started without external program directory.\r\n|");
		goto abrt;
	}

	progname=g_ptr_array_index(splitted_str,1);

	if(strchr(progname,'/')!=NULL)
	{
		g_string_sprintfa((*dest_base),"You cannot start an external program containing / in its name.\r\n|");
		goto abrt;
	}

	p=g_strconcat(ext_prog_dir,"/",progname,NULL);
	if(stat(p,&st)==-1)
	{
		g_string_sprintfa((*dest_base),"Error on %s: %s\r\n|",progname,strerror(errno));
	}
	else
	{
		/*
		start_new_ext_prog(progname)
		*/
		if (find_cnx_by_nickname(progname))
			g_string_sprintfa((*dest_base),"%s is already take by an other user\r\n|",progname);
		else
			{
		add_prog_to_start(progname);
		g_string_sprintfa((*dest_base),"%s started\r\n|",progname);
	}
	}
		
	g_free(p);

	abrt:
	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}


int cnx_io_processing(CNX_ENTRY *ptr, int no_read_wait, int no_send_retry);

/***********************/
/* kick the given nick */
/***********************/

/***********************/
/* kick the given nick */
/***********************/
static int do_kick(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	if(splitted_str->len!=3)		/* 2 parameters + the command */
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
	}
	else
	{
		char *nick;
		char	*reason;
		CNX_ENTRY	*ent;

		nick=g_ptr_array_index(splitted_str,1);
		reason=g_ptr_array_index(splitted_str,2);
		/* update the kicking user message */
		ent = find_cnx_by_nickname(nick);
		if (ent)
			g_string_sprintfa((*dest_base),"%s kicked (%s)|",nick, inet_ntoa(ent->ip));
		else
		{
			g_string_sprintfa((*dest_base),"%s not found|",nick);
			return(0);
		}

		if (strcmp("...", reason))
		{
			GString *out;
			GString_ref *out2;
			out=g_string_new("");

			/* broadcast the kicking message */
			g_string_sprintf(out,"<%s> %s is kicking %s because: %s|",
											ptr->user_nick->str,
											ptr->user_nick->str,
											nick,
											(char*)g_ptr_array_index(splitted_str,2));
			out2=conv_gstring_to_gstring_ref(out);
			queue_to_everyone(out2);
			g_string_ref_free(out2);
		}
		/* kick the user */
		kick_named_user(ptr,nick);
	}
	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}

static int do_ban_kick(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	if(splitted_str->len!=3)		/* 2 parameters + the command */
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
	}
	else
	{
		char	*nick;
		char	*reason;
		CNX_ENTRY	*ent;

		nick=g_ptr_array_index(splitted_str,1);
		reason=g_ptr_array_index(splitted_str,2);
		/* update the kicking user message */
		ent = find_cnx_by_nickname(nick);

		if (ent)
			g_string_sprintfa((*dest_base),"%s kicked (%s)|",nick, inet_ntoa(ent->ip));
		else
		{
			g_string_sprintfa((*dest_base),"%s not found|",nick);
			return(0);
		}
		if (strcmp("...", reason))
		{
			GString *out;
			GString_ref *out2;
			out=g_string_new("");

			/* broadcast the kicking message */
			g_string_sprintf(out,"<%s> %s is kicking %s because: %s|",
				 	ptr->user_nick->str,
				 	ptr->user_nick->str,
				 	nick,
				 	(char*)g_ptr_array_index(splitted_str,2));
			out2=conv_gstring_to_gstring_ref(out);
			queue_to_everyone(out2);
			g_string_ref_free(out2);
		}
		if ((ent->sock_fd!=-1) || (ent->server_id != 0))
		{
			add_tos_entry_v1_uniq(TBAN_TOSKEY,5*365*24*3600,(void*)(&(ent->ip)),sizeof(struct in_addr),NULL,0);
			send_ban_entry_to_hub(ent, 5*365*24*3600);
			if (ent->server_id == 0)
			{
				cnx_io_processing(ent,1,1);		/* flush queued message but only if possible */		
				shutdown(ent->sock_fd,0);		/* we don't want to receive more data from this connexion */
			}
			else
			{
				send_kick_entry_to_hub(ent);
			}

			{
				GString *str;
				GString_ref *str2;
			
				str=g_string_new("");
				g_string_sprintf(str,"<Hub-Security> The user %s was Banned & kicked by %s. IP: %s|",
					 	nick,ent->user_nick->str,inet_ntoa(ent->ip));
				str2 = conv_gstring_to_gstring_ref(str);
				queue_to_everyone(str2);
				g_string_ref_free(str2);
			}
		}
	}
	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}


static int do_silence_kick(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	if(splitted_str->len!=3)		/* 2 parameters + the command */
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
	}
	else
	{
		char	*nick;
		char	*reason;
		int temp_ban;
		CNX_ENTRY	*ent;

		nick=g_ptr_array_index(splitted_str,1);
		reason=g_ptr_array_index(splitted_str,2);
		/* update the kicking user message */
		ent = find_cnx_by_nickname(nick);
		if (ent)
			g_string_sprintfa((*dest_base),"%s kicked (%s)|",nick, inet_ntoa(ent->ip));
		else
		{
			g_string_sprintfa((*dest_base),"%s not found|",nick);
			return(0);
		}

		G_LOCK(conf_file);
		if(!e_db_int_get(conf_file,"KICKTEMPBAN",&temp_ban))
			temp_ban=0;
		G_UNLOCK(conf_file);
		if(ent!=NULL)
		{
			if(ent->sock_fd!=-1)
			{
				cnx_io_processing(ent,1,1);		/* flush queued message but only if possible */		
				if(temp_ban!=0)
				{
					add_tos_entry_v1_uniq(TBAN_TOSKEY,temp_ban*60,(void*)(&(ent->ip)),sizeof(struct in_addr),NULL,0);
					send_ban_entry_to_hub(ent,temp_ban*60);		 
				}
				shutdown(ent->sock_fd,0);		/* we don't want to receive more data from this connexion */
			}
			else if (ent->server_id)
			{
				if(temp_ban!=0)
				{
					add_tos_entry_v1_uniq(TBAN_TOSKEY,temp_ban*60,(void*)(&(ent->ip)),sizeof(struct in_addr),NULL,0);
					send_ban_entry_to_hub(ent,temp_ban*60);
				}
				send_kick_entry_to_hub(ent);
			}
		}
	}
	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}

static int do_low_kick(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	if(splitted_str->len!=3)		/* 2 parameters + the command */
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
	}
	else
	{
		char	*nick;
		char	*reason;
		CNX_ENTRY	*ent;
		
		nick=g_ptr_array_index(splitted_str,1);
		reason=g_ptr_array_index(splitted_str,2);
		/* update the kicking user message */
		ent = find_cnx_by_nickname(nick);
		if (ent)
			g_string_sprintfa((*dest_base),"%s kicked (%s)|",nick, inet_ntoa(ent->ip));
		else
		{
			g_string_sprintfa((*dest_base),"%s not found|",nick);
			return(0);
		}
		if(ent!=NULL)
		{			
			if(ent->sock_fd!=-1)
			{
				cnx_io_processing(ent,1,1);		/* flush queued message but only if possible */		
				shutdown(ent->sock_fd,0);		/* we don't want to receive more data from this connexion */
			}
			else
			{
				if (ent->server_id)
					send_kick_entry_to_hub(ent);
			}
		}
	}
	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}

/***************************/
/* redirect the given nick */
/***************************/
static int do_redir(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	if(splitted_str->len!=3)		/* 2 parameters + the command */
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
	}
	else
	{
		GString *rd_msg;
		char *rd;
		char *nm;
		rd_msg=g_string_new("");

		nm=g_ptr_array_index(splitted_str,1);
		rd=g_ptr_array_index(splitted_str,2);

		/* first the message to display */
		g_string_sprintf(rd_msg,"<Hub-Security> You are being redirected to %s.|",rd);
		send_str_to_named_user(nm,rd_msg->str);

		/* and then the redirection command */
		g_string_sprintf(rd_msg,"$ForceMove %s|",rd);
		send_str_to_named_user(nm,rd_msg->str);

		g_string_free(rd_msg,TRUE);

		g_string_sprintfa((*dest_base),"%s redirected\r\n|",nm);
	}

	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}

/**********************/
/* redirect all users */
/**********************/
static int do_redirall(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	if(splitted_str->len != 2) {	/* 1 parameter + the command */
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
	}
	else
	{
		GString *rd_msg;
		GString_ref *rd_msg2;
		char *rd;
		rd_msg=g_string_new("");

		rd=g_ptr_array_index(splitted_str,1);

		/* first the message to display */
		g_string_sprintf(rd_msg,"<Hub-Security> You are being redirected to %s.|",rd);
		rd_msg2 = conv_gstring_to_gstring_ref(rd_msg);
		queue_to_everyone(rd_msg2);
		g_string_ref_free(rd_msg2);


		rd_msg=g_string_new("");
		/* and then the redirection command */
		g_string_sprintf(rd_msg,"$ForceMove %s|",rd);
		rd_msg2 = conv_gstring_to_gstring_ref(rd_msg);
		queue_to_everyone(rd_msg2);
		g_string_ref_free(rd_msg2);
		g_string_sprintfa((*dest_base),"All users has been redirected\r\n|");
	}
	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}


static int do_local_redirall(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	if(splitted_str->len != 2) {	/* 1 parameter + the command */
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
	}
	else
	{
		GString *rd_msg;
		GString_ref *rd_msg2;
		char *rd;
		rd_msg=g_string_new("");

		rd=g_ptr_array_index(splitted_str,1);

		/* first the message to display */
		g_string_sprintf(rd_msg,"<Hub-Security> You are being redirected to %s.|",rd);
		rd_msg2 = conv_gstring_to_gstring_ref(rd_msg);
		queue_to_local_everyone(rd_msg2);
		g_string_ref_free(rd_msg2);
		rd_msg=g_string_new("");
		/* and then the redirection command */
		g_string_sprintf(rd_msg,"$ForceMove %s|",rd);
		rd_msg2 = conv_gstring_to_gstring_ref(rd_msg);
		queue_to_everyone(rd_msg2);
		g_string_ref_free(rd_msg2);
		g_string_sprintfa((*dest_base),"All users has been redirected\r\n|");
	}
	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}

/****************************/
/* send a broadcast message */
/****************************/

static int do_bcast(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	if(splitted_str->len!=2)		/* 1 parameters + the command */
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
	}
	else
	{
		int i;
		GString *out;
		GString_ref *out2;
		for(i=0;i<HASH_SIZE;i++)
		{
			CNX_ENTRY *ent ;

			for (ent=cnx_lst->entry[i]; ent; ent=ent->next)
			{
				out=g_string_new("");
				g_string_sprintf(out,"$To: %s From: Hub-Mass-Message $<Hub-Mass-Message> %s|", ent->user_nick->str, (char*)g_ptr_array_index(splitted_str,1));
				out2 = conv_gstring_to_gstring_ref(out);
				send_str_to(ent,out2);
				g_string_ref_free(out2);
			}
		}
		g_string_sprintfa((*dest_base),"Broadcast done\r\n|");
	}
	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}



/****************************************************/
/* add a temporary/permanent ban on the given ip(s) */
/****************************************************/
static int do_banip(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	unsigned long key=(unsigned long)xtra_param;
	time_t duration;

	if(key==TBAN_TOSKEY)
		duration=7*24*3600;				/* ban duration: 7day */
	else
		duration=3*28*24*3600;		/* ban duration: 3 month */

	if(splitted_str->len<2)		/* at least 1 parameters + the command */
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
	}
	else
	{
		int i;
		const char *ip;

		for(i=1;i<splitted_str->len;i++)
		{
			CNX_ENTRY		*ent;
			struct sockaddr_in addr_in;
			
			ip=g_ptr_array_index(splitted_str,i);
			if(ip==NULL)
				continue;

			if(inet_aton(ip,&(addr_in.sin_addr)) != 0)	/* take the string as an IP*/
 			{
				add_tos_entry_v1_uniq(key,duration,(void*)&(addr_in.sin_addr),sizeof(struct in_addr),NULL,0);
	
				/* fix me add brcastt ban addr to the cluster */
				/* send_banip_entry_to_hub(struct sockaddr_in , duration); */
				
				g_string_sprintfa((*dest_base),"ban IP %s\r\n",ip);
				continue;
			}
			ent = find_cnx_by_nickname(ip);			/* take the string as an nickname*/
			if(ent == NULL)
			{
				g_string_sprintfa((*dest_base),"'%s' is not a valid IP or not a current username.\r\n",ip);
				continue;
			}
			add_tos_entry_v1_uniq(key, duration, (void*)&(ent->ip), sizeof(struct in_addr), NULL, 0);
			send_ban_entry_to_hub(ent, duration);
			g_string_sprintfa((*dest_base),"ban IP %s\r\n",ip);
		}
		g_string_sprintfa((*dest_base),"Ban done\r\n|");
	}

	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}

/*******************************************************/
/* delete a temporary/permanent ban on the given ip(s) */
/*******************************************************/
static int do_unbanip(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	unsigned long key=(unsigned long)xtra_param;

	if(splitted_str->len<2)		/* at least 1 parameters + the command */
	{
		g_string_sprintfa((*dest_base),"invalid number of parameters for %s\r\n|",cmd);
	}
	else
	{
		int i;
		const char *ip;

		for(i=1;i<splitted_str->len;i++)
		{
			ip=g_ptr_array_index(splitted_str,i);
			if(ip!=NULL)
			{
				struct in_addr inp;

				if(inet_aton(ip,&inp)==0)
				{
					g_string_sprintfa((*dest_base),"Fail to unban IP '%s'. Invalid address\r\n",ip);
				}
				else
				{
					delete_this_tos_entry(key,(void*)&inp,sizeof(inp),0);
					g_string_sprintfa((*dest_base),"unban IP %s\r\n",ip);
				}
			}
		}

		g_string_sprintfa((*dest_base),"Unban done\r\n|");
	}

	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}

static int append_ban_entry(const time_t timeout, const char *ptr1, const int len1, const char *ptr2, const int len2, void *xtra)
{
	GString **dest_base=xtra;

	if((ptr1!=NULL)&&(len1==sizeof(struct in_addr)))
	{
		struct in_addr *inp=(void*)ptr1;
		struct tm tm;
		char timestamp[512];

		localtime_r(&timeout,&tm);
		strftime(timestamp,sizeof(timestamp),"%c",&tm);

		/* inet_ntoa is not thread safe, it works with a static buffer, we will do the conversion manually */
		g_string_sprintfa((*dest_base),"IP: %u.%u.%u.%u	ban until %s\r\n",
								(unsigned)((ntohl(inp->s_addr)>>24)&255),
								(unsigned)((ntohl(inp->s_addr)>>16)&255),
								(unsigned)((ntohl(inp->s_addr)>>8)&255),
								(unsigned)((ntohl(inp->s_addr))&255),
								timestamp
								);
	}
	return 0;	/* continue */
}

/********************************************/
/* print the temporary/permanently ban list */
/********************************************/
static int do_banlist(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	unsigned long key=(unsigned long)xtra_param;

	if(key==TBAN_TOSKEY)
		g_string_sprintfa((*dest_base),"Tban list:\r\n");
	else
		g_string_sprintfa((*dest_base),"Pban list:\r\n");

	scan_all_tos_entry(key,append_ban_entry,dest_base);

	g_string_sprintfa((*dest_base),"|");
	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}

typedef struct
{
	 const char *cmd;
	 size_t cmd_len;
	 int (*fnc)(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, 
					const GString *str, char *xtra_param, const GPtrArray *splitted_str);
			/* function to call. (ERROR: the following sentence is FALSE => This function MUST free str when it is not useful anymore) */
			/* the function must return >=0 if the connection must remain alive (no error (0) or */
			/* recoverable error (>0)) and a value <0 if there is a fatal error) */
			/* *dest_base is a the beginning of the reply to send. To create the full reply, */
			/* add the message you want to send and ends the string with a |. Use send_const_str_to(ptr,(*dest_base)->str) */
			/* to send the reply */
	 char *xtra_param;
	int op_level;		/* it is a bit mask. If (1<<is_op) is set, the user can use this command */
} KNOWN_CMD;


static KNOWN_CMD hub_cmd[]={
	{"-help",		sizeof("-help")-1,		do_help,		NULL, OP_PRIV},

	/* user account management */
	{"-ulist",		sizeof("-ulist")-1,		do_ulist,		NULL,MASTER_PRIV},
	{"-add ",		sizeof("-add ")-1,		do_add,			NULL,MASTER_PRIV},
	{"-del ",		sizeof("-del ")-1,		do_del,			NULL,MASTER_PRIV},
	{"-rename ",		sizeof("-rename ")-1,		do_rename,		NULL,MASTER_PRIV},
	{"-passwd ",		sizeof("-passwd ")-1,		do_passwd,		NULL,MASTER_PRIV},
	{"-type ",		sizeof("-type ")-1,		do_type,		NULL,MASTER_PRIV},

	/* database management */
	{"-dbshow ",		sizeof("-dbshow ")-1,		do_dbshow,		NULL,MASTER_PRIV},
	{"-dbadd ",		sizeof("-dbadd ")-1,		do_dbadd,		NULL,MASTER_PRIV},
	{"-dbset ",		sizeof("-dbset ")-1,		do_dbset,		NULL,MASTER_PRIV},
	{"-dbdel ",		sizeof("-dbdel ")-1,		do_dbdel,		NULL,MASTER_PRIV},
	/* must be the last of -db */
	{"-db",			sizeof("-db")-1,		do_db,			NULL,MASTER_PRIV},
							
	/* external program start */
	{"-startprg ",		sizeof("-startprg ")-1,		do_startprg,		NULL,MASTER_PRIV},

	/* operator command */
	{"-kick ",		sizeof("-kick ")-1,		do_kick,		NULL,OP_PRIV},
	{"-lkick ",		sizeof("-lkick ")-1,		do_low_kick,		NULL,OP_PRIV},
	{"-skick ",		sizeof("-skick ")-1,		do_silence_kick,	NULL,OP_PRIV},
	{"-bankick ",		sizeof("-bankick ")-1,		do_ban_kick,		NULL,OP_PRIV},
	{"-redir ",		sizeof("-redir ")-1,		do_redir,		NULL,OP_PRIV},
	{"-redirall ",		sizeof("-redirall ")-1,		do_redirall,		NULL,MASTER_PRIV},
	{"-rediralllocal ",	sizeof("-rediralllocal ")-1,	do_local_redirall,	NULL,MASTER_PRIV},
	{"-bcast ",		sizeof("-bcast ")-1,		do_bcast,		NULL,OP_PRIV},
	{"-tbanip ",		sizeof("-tbanip ")-1,		do_banip,		(void*)TBAN_TOSKEY,OP_PRIV},
	{"-tunbanip ",		sizeof("-tunbanip ")-1,		do_unbanip,		(void*)TBAN_TOSKEY,OP_PRIV},
	{"-tbanlist",		sizeof("-tbanlist")-1,		do_banlist,		(void*)TBAN_TOSKEY,OP_PRIV},
	{"-pbanip ",		sizeof("-pbanip ")-1,		do_banip,		(void*)PBAN_TOSKEY,OP_PRIV},
	{"-punbanip ",		sizeof("-punbanip ")-1,		do_unbanip,		(void*)PBAN_TOSKEY,OP_PRIV},
	{"-pbanlist",		sizeof("-pbanlist")-1,		do_banlist,		(void*)PBAN_TOSKEY,OP_PRIV},
#ifdef DEBUG
	{"-dump ",		sizeof("-dump ")-1,		do_dump,		NULL,MASTER_PRIV},
	{"-dumpqueue",		sizeof("-dumpqueue")-1,		do_dump_queue,		NULL,MASTER_PRIV},
#endif
	{"-revip ",		sizeof("-revip ")-1,		do_rev_ip,		NULL,OP_PRIV},
	{"-getip ",		sizeof("-getip ")-1,		do_get_ip,		NULL,OP_PRIV},
						
	{"-setfusionuser ",	sizeof("-setfusionuser ")-1,	do_set_fusr,		NULL,FUSION_PRIV},
	{"-setfusionpass ",	sizeof("-setfusionpass ")-1,	do_set_fpass,		NULL,FUSION_PRIV},
	{"-fusion ",		sizeof("-fusion ")-1,		do_fusion,		NULL,FUSION_PRIV},
	{"-ufusion",		sizeof("-ufusion")-1,		do_unfusion,		NULL,FUSION_PRIV},
	{"-set ",		sizeof("-set ")-1,		do_set,			NULL,MASTER_PRIV},
	{"-reload",		sizeof("-reload")-1,		do_reload,		NULL,MASTER_PRIV},
	{NULL,0,NULL,NULL},
};

/*********************************************************************************/
/* run the given command. Incoming string is splitted, access rights are checked */
/*********************************************************************************/
static int run_hub_cmd(KNOWN_CMD *hcmd, CNX_ENTRY *ptr, GString **dest, const GString *inp)
{
	if ((hcmd->op_level == 0) || ((ptr->is_op & hcmd->op_level) != 0))
	{
		GStringChunk *chunk=NULL;
		GPtrArray *tbl=NULL;
		int ret=0;

		split_string_into_array(&tbl,&chunk,inp->str," ");
		/* now, we can process the command */
		if((chunk!=NULL)&&(tbl!=NULL))
		{
			ret=(hcmd->fnc)(hcmd->cmd,ptr,dest,inp,hcmd->xtra_param,tbl);
		}

		/* free allocated data */
		if(tbl!=NULL)
			g_ptr_array_free(tbl,TRUE);
		if(chunk!=NULL)
			g_string_chunk_free(chunk);
		return ret;
	}
	else
	{
		/* not enough rights */
		g_string_sprintfa((*dest),"You are not allowed to use this command.|");
			send_const_str_to(ptr,(*dest)->str);
		return 0;
	}
}

/*************************************************************************/
/* the given string has been sent to "Hub-Security", it may be a command */
/* This function processes the string and performs the wanted action		 */
/*************************************************************************/
/* output: =0: ok																*/
/*				 <0: fatal error (=> abort connection) */
/*				 >0: recoverable error (continue)			*/
/*************************************************/
int process_hub_command(CNX_ENTRY *ptr,char *string)
{
	int ret=0;
	GString *inp;
	GString *dest=NULL;
	char *t;
	int i;
	int fnd=0;

	inp=g_string_new(string);

	/* printf("%s\n",inp->str); */

	t=strchr(inp->str,' ');
	SKIP_SPACE(t);

	if(*t=='\0')
	{
		ret=1;
		goto eoc;		/* non fatal error: nothing entered */
	}

	/* remote trash before the command */
	inp=g_string_erase(inp,0,t-inp->str);
	if(inp->len>1)
		inp=g_string_truncate(inp,inp->len-1);		/* remove trailing | */

	/* printf("%s\n",inp->str); */
	/* build the beginning of the query reply */
	dest=g_string_new("");
	g_string_sprintf(dest,"$To: %s From: Hub-Security $",ptr->user_nick->str);

	i=0;
	while(hub_cmd[i].cmd!=NULL)
	{
		if(!strncmp(inp->str,hub_cmd[i].cmd,hub_cmd[i].cmd_len))
		{
			fnd=1;
			ret=run_hub_cmd(&hub_cmd[i],ptr,&dest,inp);

			break;
		}
		i++;
	}

	if(!fnd)
	{
		g_string_sprintfa(dest,"Unknown command: %s|",inp->str);
		send_const_str_to(ptr,dest->str);
	}
	eoc:
	if(dest!=NULL)
		g_string_free(dest,TRUE);
	g_string_free(inp,TRUE);

	return 0;
}

/*****************************************/
/* display the list of all user commands */
/*****************************************/
static int do_user_help(const char *cmd, CNX_ENTRY *ptr, GString **dest_base, const GString *str, char *xtra_param,const GPtrArray *splitted_str)
{
	g_string_sprintfa((*dest_base),"** User command list:\r\n"
											 "\r\n"
											 "+help\r\n"
											 "               display this help\r\n"
										 	 "\r\n"
											 "+new_mchat chatname\r\n"
											 "               create a new private multi-users chat named \"chatname\". The\r\n"
											 "               private multi-users chat (mchat in short) works like the normal\r\n"
											 "               private chat except you can be more than 2 users in this chat.\r\n"
											 "               Note: it is not possible to create to mchat with the same name.\r\n"
											 "\r\n"
							 				 "+invite_mchat chatname nick1 [nick2....]\r\n"
											 "               invite the given users to the mchat named \"chatname\".\r\n"
											 "               Note: only someone inside the mchat can invite someone outside\r\n"
											 "               the mchat inside.\r\n"
											 "\r\n"
											 "+list\r\n"
											 "               list the existing irc-like chan.\r\n"
											 "\r\n"
											 "+join channame\r\n"
											 "               connect to a public irc like chan named \"channame\".\r\n"
											 "\r\n"
											 "+leave channame\r\n"
											 "               leave to a public irc like chan named \"channame\".\r\n"
											 "\r\n|");
	send_const_str_to(ptr,(*dest_base)->str);
	return 0;
}
#define EVERYONE_MSK 0
static KNOWN_CMD user_hub_cmd[]={
	{"+new_mchat ",			sizeof("+new_mchat ")-1,			create_multi_chat,		NULL,EVERYONE_MSK},
	{"+invite_mchat ",		sizeof("+invite_mchat ")-1,		invite_multi_chat,		NULL,EVERYONE_MSK},
	{"+leave_mchat ",			sizeof("+leave_mchat ")-1,			cmd_leave_multi_chat,	NULL,EVERYONE_MSK},
	{"+join ",		sizeof("+join ")-1,		join_pubchat,		NULL,EVERYONE_MSK},
	{"+leave ",		sizeof("+leave ")-1,		cmd_leave_multi_pubchat,NULL,EVERYONE_MSK},
	{"+list",		sizeof("+list")-1,		cmd_list_multi_pubchat ,NULL,EVERYONE_MSK},
	{"+help",					sizeof("+help")-1,					do_user_help,				NULL,EVERYONE_MSK},
	{NULL,0,NULL,NULL},
};

/*************************************************************************/
/* the given string has been sent to any user but always starts with a + */
/* This function processes the string and performs the wanted action     */
/*************************************************************************/
/* output: =0: ok                                */
/*         <0: fatal error (=> abort connection) */
/*         >0: recoverable error (continue)      */
/*************************************************/
int process_user_hub_command(CNX_ENTRY *ptr,char *string)
{
	int ret=0;
	GString *inp;
	GString *dest=NULL;
	char *t;
	int i;
#if 0
	int fnd=0;
#endif

	inp=g_string_new(string);

	//printf("%s\n",inp->str);

#if 0
	t=strchr(inp->str,' ');
#else
	t=inp->str;
#endif
	SKIP_SPACE(t);

	if(*t=='\0')
	{
		ret=1;
		goto eoc;		/* non fatal error: nothing entered */
	}

	/* remote trash before the command */
	inp=g_string_erase(inp,0,t-inp->str);
	if(inp->len>1)
		inp=g_string_truncate(inp,inp->len-1);		/* remove trailing | */

	//printf("%s\n",inp->str);
	/* build the beginning of the query reply */
	dest=g_string_new("");
	g_string_sprintf(dest,"$To: %s From: Hub-Security $",ptr->user_nick->str);

	i=0;
	while(user_hub_cmd[i].cmd!=NULL)
	{
		if(!strncmp(inp->str,user_hub_cmd[i].cmd,user_hub_cmd[i].cmd_len))
		{
#if 0
			fnd=1;
#endif
			ret=run_hub_cmd(&user_hub_cmd[i],ptr,&dest,inp);

			break;
		}
		i++;
	}

#if 0
	/* if the command is not found, we don't return any error message because */
	/* the command can be handled by a script */
	if(!fnd)
	{
		g_string_sprintfa(dest,"Unknown command: %s|",inp->str);
		send_str_to(ptr,dest->str);
	}
#endif
	eoc:
	if(dest!=NULL)
		g_string_free(dest,TRUE);
	g_string_free(inp,TRUE);

	return 0;
}

typedef struct
{
	const char *cmd;
	void (*fnc)(const char *evt_name, const char *evt_emitter, int nb_args, va_list ap, char *xtra_param);
	/* NOTE: all parameters are read-only because you may have more than 1 event handler for each event */
	char *xtra_param;
} KNOWN_EVENT;

static KNOWN_EVENT hub_event_hdl[]={
	{"quit",leave_multi_chat,NULL},
	{"privchat",privchat_multi_chat,NULL},
	{"getnicklist",getnicklist_multi_chat,NULL},
	{NULL,NULL,NULL},
};

/**************************************************************************************************/
/* in the same way as it exists for PERL scripts, there is an event handler for embedded commands */
/**************************************************************************************************/
/* evt_name and evt_emitter are the same as PERL scripts one */
/* ap is a va_list of nb_args (char *)                       */
/*************************************************************/
void hub_event_command(const char *evt_name, const char *evt_emitter, int nb_args, va_list ap)
{
	int i;
	i=0;
	while(hub_event_hdl[i].cmd!=NULL)
	{
		if(!strcmp(evt_name,hub_event_hdl[i].cmd))
		{
			(hub_event_hdl[i].fnc)(evt_name,evt_emitter,nb_args,ap,hub_event_hdl[i].xtra_param);
		}
		i++;
	}
}


